# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import logging
from collections import Sequence
from armet import http
from armet.resources.resource import base


logger = logging.getLogger(__name__)


class ManagedResource(base.Resource):
    """Implements the RESTful resource protocol for managed resources.

    @note
        This is not the class to derive from when implementing your own
        resources. Derive from `armet.resources.ManagedResource` (defined in
        the `__init__.py`).
    """

    class Meta:
        # The managed resource access pattern.
        # Traversal is handled a bit more dyanmically than allowed with
        # the simple pattern syntax.
        patterns = [
            r'^'
            r'(?:\:(?P<directives>[^/\(\)]*))?'
            r'(?:\((?P<query>[^/]*)\))?'
            r'(?:/(?P<slug>[^/]+?))?'
            r'(?:/(?P<path>.+?))??'
            r'(?:\.(?P<extensions>[^/]+?))??/??$'
        ]

    #! An ordered dictionary of attributes that are gathered from both
    #! the metaclass and attributes defined directly on the resource.
    #!
    #! Attributes may be declared in a few different
    #! ways. The next few code blocks are identical in function.
    #!
    #! @code
    #! from armet import resources
    #!
    #! class Resource(resources.Resource):
    #!     name = resources.Attribute()
    #!     created = resources.Attribute('created')
    #! @endcode
    #!
    #! @code
    #! from armet import resources
    #!
    #! class Resource(resources.Resource):
    #!     class Meta:
    #!         include = {
    #!             'name': resources.Attribute()
    #!             'created': resources.Attribute('created')
    #!         }
    #! @endcode
    #!
    #! @code
    #! from armet import resources
    #!
    #! class Resource(resources.Resource):
    #!     class Meta:
    #!         include = {
    #!             'name': None,
    #!             'created': 'created'
    #!         }
    #! @endcode
    #!
    #! @code
    #! from armet import resources
    #!
    #! class Resource(resources.Resource):
    #!     class Meta:
    #!         include = ('name',)
    #!
    #!     created = resources.Attribute('created')
    #! @endcode
    attributes = None

    #! The slug that is used to identify the resource an item is being
    #! requested. None if a list is being requested.
    slug = None

    @classmethod
    def parse(cls, path):
        result = super(ManagedResource, cls).parse(path)
        if result:
            # Found something; parse the result.
            resource, data, end = result

            # Normalize the list arguments.
            for sep, name in (('.', 'extensions'), (':', 'directives')):
                if data[name]:
                    data[name] = data[name].split(sep)

                else:
                    data[name] = []

            # Reset the result.
            result = resource, data, end

        # Return what we got.
        return result

    @property
    def allowed_operations(self):
        """Retrieves the allowed operations for this request."""
        if self.slug is not None:
            return self.meta.detail_allowed_operations

        return self.meta.list_allowed_operations

    def assert_operations(self, *args):
        """Assets if the requested operations are allowed in this context."""
        if not set(args).issubset(self.allowed_operations):
            raise http.exceptions.Forbidden()

    def make_response(self, data=None, status=http.client.OK):
        """Fills the response object from the passed data."""
        if data is not None:
            # Prepare the data for transmission.
            data = self.prepare(data)

            # Encode the data using a desired encoder.
            self.response.write(data, serialize=True)

        # Make sure that the status code is set.
        self.response.status = status

    def prepare(self, data):
        if data is None:
            # No data; return nothing.
            return None

        if (not isinstance(data, six.string_types)
                and isinstance(data, Sequence)):
            # Attempt to prepare each item of the iterable (as long as
            # we're not a string or some sort of mapping).
            for index, value in enumerate(data):
                data[index] = self.item_prepare(data[index])

            return data

        # Prepare just the singular value and return.
        return self.item_prepare(data)

    def item_prepare(self, item):
        # Initialize the object that hold the resultant item.
        obj = {}

        # Iterate through the attributes and build the object from the item.
        for name, attribute in self.attributes.items():
            if attribute.include:
                # Run the attribute through its prepare cycle.
                obj[name] = attribute.prepare(
                    # Optional preparation cycle on the resource.
                    self.preparers[name](
                        # Retrieves the value from the object.
                        self, item, attribute.get(item)))

        # Return the resultant object.
        return obj

    def clean(self, data):
        if not data:
            # If no data; return an empty object.
            return {}

        if (not isinstance(data, six.string_types)
                and isinstance(data, Sequence)):
            # Attempt to clean each item of the iterable (as long as
            # we're not a string or some sort of mapping).
            for index, value in enumerate(data):
                data[index] = self.item_clean(data[index])

            return data

        # Clean just the singular value and return.
        return self.item_clean(data)

    def item_clean(self, item):
        # Initialize the object that hold the resultant item.
        obj = {}

        # Iterate through the attributes and build the object from the item.
        for name, attribute in self.attributes.items():
            if attribute.include:
                # Run the attribute through its prepare cycle.
                obj[name] = self.cleaners[name](
                    # Micro preparation cycle on the attribute object.
                    self, attribute.clean(item.get(name)))

        # Return the resultant object.
        return obj

    @property
    def http_allowed_methods(self):
        if self.slug is None:
            # No slug means that we're accessing this as a list.
            return self.meta.http_list_allowed_methods

        # A slug exists; this is being accessed as an item.
        return self.meta.http_detail_allowed_methods

    def assert_http_allowed_methods(self, request):
        # No super call as the following replaces it by checking
        # only against the more specific `list_*` or `detail_*` http
        # allowed methods.

        # Check against `list_*` or `detail_*` allowed methods.
        if request.method not in self.http_allowed_methods:
            # Current method is found to not be allowed for
            # this type of access; raise.
            raise http.exceptions.MethodNotAllowed(self.http_allowed_methods)

    def get(self, request, response):
        """Processes a `GET` request."""
        # Ensure we're allowed to read the resource.
        self.assert_operations('read')

        try:
            # Delegate to `read` to retrieve the items.
            items = self.read()

        except AttributeError:
            # No read method defined.
            raise http.exceptions.NotImplemented()

        if self.slug is not None and not items:
            # Requested a specific resource but nothing is returned.
            raise http.exceptions.NotFound()

        # Build the response object.
        self.make_response(items)

    def post(self, request, response):
        """Processes a `POST` request."""
        if self.slug is not None:
            # Don't know what to do an item access.
            raise http.exceptions.NotImplemented()

        # Ensure we're allowed to create a resource.
        self.assert_operations('create')

        # Deserialize and clean the incoming object.
        data = self.clean(self.request.read(deserialize=True))

        try:
            # Delegate to `create` to create the item.
            item = self.create(data)

        except AttributeError:
            # No read method defined.
            raise http.exceptions.NotImplemented()

        # Build the response object.
        self.make_response(item, status=http.client.CREATED)

    def put(self, request, response):
        """Processes a `PUT` request."""
        if self.slug is None:
            # Mass-PUT is not implemented.
            raise http.exceptions.NotImplemented()

        try:
            # Check if the resource exists.
            target = self.read()

        except AttributeError:
            # No read method defined.
            raise http.exceptions.NotImplemented()

        # Deserialize and clean the incoming object.
        data = self.clean(self.request.read(deserialize=True))

        if target is not None:
            # Ensure we're allowed to update the resource.
            self.assert_operations('update')

            try:
                # Delegate to `update` to create the item.
                self.update(target, data)

            except AttributeError:
                # No read method defined.
                raise http.exceptions.NotImplemented()

            # Build the response object.
            self.make_response(target, status=http.client.OK)

        else:
            # Ensure we're allowed to create the resource.
            self.assert_operations('create')

            try:
                # Delegate to `create` to create the item.
                target = self.create(data)

            except AttributeError:
                # No read method defined.
                raise http.exceptions.NotImplemented()

            # Build the response object.
            self.make_response(target, status=http.client.CREATED)

    def delete(self, request, response):
        """Processes a `DELETE` request."""
        if self.slug is None:
            # Mass-DELETE is not implemented.
            raise http.exceptions.NotImplemented()

        # Ensure we're allowed to destroy a resource.
        self.assert_operations('destroy')

        try:
            # Delegate to `destroy` to destroy the item.
            self.destroy()

        except AttributeError:
            # No read method defined.
            raise http.exceptions.NotImplemented()

        # Build the response object.
        self.make_response(status=http.client.NO_CONTENT)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet.http import client


class BaseHTTPException(BaseException):
    status = None

    def __init__(self, content=None, headers=None):
        #! Body of the exception message.
        self.content = content

        #! Additional headers to place with the response.
        self.headers = headers or {}


class BadRequest(BaseHTTPException):
    status = client.BAD_REQUEST


class Forbidden(BaseHTTPException):
    status = client.FORBIDDEN


class NotFound(BaseHTTPException):
    status = client.NOT_FOUND


class MethodNotAllowed(BaseHTTPException):
    status = client.METHOD_NOT_ALLOWED

    def __init__(self, allowed):
        super(MethodNotAllowed, self).__init__(
            headers={'Allow': ', '.join(allowed)})


class NotAcceptable(BaseHTTPException):
    status = client.NOT_ACCEPTABLE


class UnsupportedMediaType(BaseHTTPException):
    status = client.UNSUPPORTED_MEDIA_TYPE


class InternalServerError(BaseHTTPException):
    status = client.INTERNAL_SERVER_ERROR


class NotImplemented(BaseHTTPException):
    status = client.NOT_IMPLEMENTED



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .resource import Resource
from .managed import ManagedResource
from .model import ModelResource
from .attributes import (Attribute,
                         BooleanAttribute,
                         TextAttribute,
                         IntegerAttribute,
                         UUIDAttribute,
                         DateAttribute,
                         TimeAttribute,
                         DateTimeAttribute,
                         TimezoneAttribute)

__all__ = [
    'Resource',
    'ManagedResource',
    'ModelResource',
    'Attribute',
    'BooleanAttribute',
    'TextAttribute',
    'IntegerAttribute',
    'UUIDAttribute',
    'DateAttribute',
    'TimeAttribute',
    'DateTimeAttribute',
    'TimezoneAttribute'
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import collections
import six
import uuid
import datetime
from armet import exceptions


class Attribute(object):
    """Generic attribute; (de)serialzied as text.
    """

    #! Underyling type of the attribute.
    type = None

    def __init__(self, path=None, **kwargs):
        """Initializes this attribute with the given properties."""
        #! If this attribute can be read via direct
        #! access (eg. GET /resource/1/attribute)
        #! @warning Not implemented.
        self.readable = kwargs.get('readable', True)

        #! If this attribute can be written via direct
        #! access (eg. PUT /resource/1/attribute) or modification of the
        #! resource (eg. PUT /resource/1 or PATCH /resource/1).
        #! @warning Not implemented.
        self.writable = kwargs.get('writable', True)

        #! If this attribute is included in the resource body.
        #! @warning Not implemented.
        self.include = kwargs.get('include', True)

        #! If this attribute can accept null as a value.
        #! @warning Not implemented.
        self.null = kwargs.get('null', True)

        #! If this attribute must have some value.
        #! @warning Not implemented.
        self.required = kwargs.get('required', True)

        #! If this attribute is represented as a collection (aka. array).
        self.collection = kwargs.get('collection', False)

        #! The path reference of where to find this attribute on an
        #! item (eg. 'name' references the name key if the read method returns
        #! a dictionary.)
        #!
        #! The path may be dot-separated to indicate simple traversal.
        #! Eg. user.name could be obj['user'].name
        self.path = path

        if self.path:
            # Explode the path into segments.
            self._segments = path.split('.')

            # Initialize the getters and setters.
            self._getters = []
            self._setter = None

    def clone(self):
        # Construct a new this.
        return self.__class__(**self.__dict__)

    def _make_getter(self, path, cls, instance):
        # Attempt to get an unbound class property
        # that may be a descriptor.
        obj = getattr(cls, path, None)
        if obj is not None:
            if hasattr(obj, '__call__'):
                # The descriptor is callable.
                return obj.__call__

            if hasattr(obj, '__get__'):
                # This has a data descriptor.
                return lambda o, c=cls, g=obj.__get__: g(o, c)

        else:
            # Check for another kind of descriptor.
            descriptor = cls.__dict__.get(path)
            if descriptor and hasattr(descriptor, '__get__'):
                return descriptor.__get__

        if issubclass(cls, collections.Mapping):
            return lambda o, n=path: o.get(n)

        # No alternative; let's pretend this will work (which it will
        # most of the time).
        return lambda o, n=path: o.__dict__.get(n)

    def _make_setter(self, path, cls, instance):
        # Attempt to get an unbound class property
        # that may be a descriptor.
        obj = getattr(cls, path, None)
        if obj is not None:
            if hasattr(obj, '__set__'):
                # This has a data descriptor.
                return lambda o, v, s=obj.__set__: s(o, v)

        else:
            # Check for another kind of descriptor.
            descriptor = cls.__dict__.get(path)
            if descriptor and hasattr(descriptor, '__set__'):
                return descriptor.__set__

        if issubclass(cls, collections.Mapping):
            def setter(o, v, n=path):
                o[n] = v

            return setter

        # No alternative; let's pretend this will work (which it will
        # most of the time).
        def setter(o, v, n=path):
            o.__dict__[n] = v

        return setter

    def get(self, target, parent=False):
        """Retrieves the value of this attribute from the passed object."""
        if not self.path:
            # If we do not have a path; we cannot automatically
            # resolve our value; return nothing.
            return None

        for getter in self._getters[:-1 if parent else None]:
            # Iterate and resolve the attribute path.
            target = getter(target)

        if self._segments and target is not None:
            # Value isn't none and we still have additional segments left
            # to resolve into getters.
            while self._segments:
                if target is None:
                    # We no longer have a value to use to attempt
                    # to resolve additional segments; bail for now.
                    break

                # Remove any path segments that have been resolved
                # into getters.
                segment = self._segments.pop(0)

                # Build the getter corresponding to this path
                # segment.
                getter = self._make_getter(
                    segment, target.__class__, target)

                # Append the getter.
                self._getters.append(getter)

                if self._segments or not parent:
                    # Utilize the getter now.
                    target = getter(target)

        # Return what has been accessed.
        return target

    def set(self, target, value):
        """Stores the value on the passed target."""
        if not self.path:
            # If we do not have a path; we cannot automatically
            # resolve our value; do nothing.
            return

        # Iterate and resolve the parent of our target.
        parent = self.get(target, parent=True)

        if self._setter is None:
            # Resolve our setter if needed.
            self._setter = self._make_setter(
                self.path.split('.')[-1], parent.__class__, parent)

        # Set the target.
        self._setter(parent, value)

    def prepare(self, value):
        """Prepares the value for serialization."""
        return value

    def clean(self, value):
        """Cleans the value in preparation for deserialization."""
        return value

    def try_clean(self, value):
        """Cleans the value in preparation for deserialization.

        Does not throw errors; returns None if the conversion failed to
        happen.
        """
        try:
            return self.clean(value)

        except ValueError:
            return None


class TextAttribute(Attribute):
    """Represents text.
    """

    type = str

    def prepare(self, value):
        """Stringifies the value."""
        return str(value) if value is not None else None


class BooleanAttribute(Attribute):
    """
    Represents a boolean; prepared as a python bool and cleaned as
    an actual bool in most deserializers.
    """

    type = bool

    #! Textual values accepted for `True`.
    TRUE = (
        'true',
        't',
        'yes',
        'y',
        'on',
        '1'
    )

    #! Textual values accepted for `False`.
    FALSE = (
        'false',
        'f',
        'no',
        'n',
        'off',
        '0'
    )

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        if value is True or value is False:
            # Value is a python boolean; just return it.
            return value

        if value.strip().lower() in self.TRUE:
            # Some sort of truthy value.
            return True

        if value.strip().lower() in self.FALSE:
            # Some sort of falsy value.
            return False

        # Neither true or false matches; return a boolifyed version of
        # whatever we have.
        return bool(value)


class IntegerAttribute(Attribute):
    """Represents an integer; serialzied as a python integer.
    """

    type = int

    def clean(self, value):
        if isinstance(value, six.string_types):
            # Strip the string of whitespace
            value = value.strip()

        if value is None:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to coerce whatever we have as an int.
            return int(value)

        except ValueError:
            # Failed to do so.
            raise ValueError('Not a valid integer.')


try:
    # Attempt to import and make use of date/time libraries.
    # TODO: Make note of this optional dep somewhere.
    from dateutil.parser import parse as parse_datetime
    from time import mktime

except ImportError:
    # No support for date/times.
    parse_datetime = None


class _TemporalAttribute(object):
    """Represents a temporal attribute, such as a date or time.
    """

    def __init__(self, *args, **kwargs):
        # Ensure we have support.
        if parse_datetime is None:
            raise exceptions.ImproperlyConfigured(
                'Use of temporal attributes requires at '
                'least python-dateutil (and optionally parsedatetime).')

        # Continue on.
        super(_TemporalAttribute, self).__init__(*args, **kwargs)

    def prepare(self, value):
        # Serialize as ISO format.
        return value.isoformat()

    def clean(self, value):
        if not value:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to use the dateutil library to parse.
            return parse_datetime(value, fuzzy=False)

        except (ValueError, AttributeError):
            # Not a strictly formatted date; return nothing.
            pass

        try:
            # Attempt to magic a date out of it.
            # TODO: List this somewhere as an optional dep.
            from parsedatetime import parsedatetime as pdt
            c = pdt.Constants()
            c.BirthdayEpoch = 80  # TODO: Figure out what this is.
            p = pdt.Calendar(c)
            result = p.parse(value)
            if result[1] != 0:
                return datetime.datetime.fromtimestamp(mktime(result[0]))

        except (NameError, ImportError, TypeError):
            # No magical date/time support.
            pass

        # Couldn't figure out what we're dealing with.
        raise ValueError('Invalid date/time or in an invalid format.')


class DateAttribute(_TemporalAttribute, Attribute):

    type = datetime.date

    def clean(self, value):
        value = super(DateAttribute, self).clean(value)
        if value:
            # Constrain to just the date part.
            value = value.date()
        return value


class TimeAttribute(_TemporalAttribute, Attribute):

    type = datetime.time

    def clean(self, value):
        value = super(TimeAttribute, self).clean(value)
        if value:
            # Constrain to just the date part.
            value = value.time()
        return value


class DateTimeAttribute(_TemporalAttribute, Attribute):
    type = datetime.datetime


class UUIDAttribute(Attribute):

    type = uuid.UUID

    def prepare(self, value):
        # Serialize as the 16-digit hex representation.
        return value.hex if value else value

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to coerce the UUID.
            return uuid.UUID(value)

        except ValueError:
            raise ValueError(
                'UUID must be of the form: '
                'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx')


try:
    # TODO: List this somewhere as an optional dep.
    import pytz

except ImportError:
    # No timezone support.
    pytz = None


class TimezoneAttribute(Attribute):

    if pytz is not None:
        type = pytz.tzfile.DstTzInfo

    def __init__(self, *args, **kwargs):
        # Ensure we have support.
        if pytz is None:
            raise exceptions.ImproperlyConfigured(
                'Use of the timezone attribute requires pytz')

        # Continue on.
        super(TimezoneAttribute, self).__init__(*args, **kwargs)

    def prepare(self, value):
        return value.zone

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to coerce the timezone.
            return pytz.timezone(value)

        except pytz.UnknownTimeZoneError:
            raise ValueError('Unknown time zone.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import json
from collections import Iterable
from .base import Serializer
from armet import media_types


class JSONSerializer(Serializer):

    media_types = media_types.JSON

    @staticmethod
    def _default(obj):
        if isinstance(obj, Iterable):
            # This is an iterable but not recognizable as such
            # by the JSON serializer (eg. generator).
            return list(obj)

        # Raise up our hands; we cannot serialize this.
        raise ValueError('Unable to serialize {}'.format(obj))

    def __init__(self, *args, **kwargs):
        # Let the base class figure out things.
        super(JSONSerializer, self).__init__(*args, **kwargs)

        #! The configuration options that are passed to the
        #! JSON serializer.
        self.options = {
            'ensure_ascii': True,
            'separators': (',', ':',),
            'default': self._default
        }

    def serialize(self, obj=None):
        # If we have nothing; serialize as an empty object.
        if obj is None:
            obj = {}

        # Ensure it is atleast wrapped in an array.
        if isinstance(obj, six.string_types) or not isinstance(obj, Iterable):
            obj = [obj]

        # Serialize the resultant text.
        text = json.dumps(obj, **self.options)

        # Return us to the base to enclose it inside of a response object.
        return super(JSONSerializer, self).serialize(text)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class ImproperlyConfigured(BaseException):
    """Something has been set up or configured incorrectly.
    """


class InvalidOperation(BaseException):
    """
    Something is being asked to operate outside of the defined specification
    of it operating.
    """



# -*- coding: utf-8 -*-
from armet.utils import classproperty


def connect(name):
    """Creates a proxy to a connector for a named attribute or method.
    """

    @property
    def proxy(self):
        for connector in self.connectors:
            data = vars(connector)
            if name in data:
                return data[name].__get__(self)

        raise AttributeError("class object '%s' has no attribute '%s'" % (
            type(self).__name__, name))

    return proxy


def classconnect(name):
    """Creates a proxy to a connector for a named attribute or classmethod.
    """

    @classproperty
    def proxy(cls):
        for connector in cls.connectors:
            data = vars(connector)
            if name in data:
                return data[name].__get__(cls)

        raise AttributeError("type object '%s' has no attribute '%s'" % (
            cls.__name__, name))

    return proxy



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import sys
import platform


# Required test dependencies.
test_dependencies = (
    # Test runner.
    'pytest',

    # Ensure PEP8 conformance.
    'pytest-pep8',

    # Ensure test coverage.
    'pytest-cov',

    # Installs a WSGI application that intercepts requests made to a hostname
    # and port combination for testing.
    'wsgi_intercept == 0.6.0',

    # HTTP request abstraction layer over httplib.
    'httplib2',

    # The Web framework for perfectionists with deadlines.
    'django',

    # A microframework based on Werkzeug, Jinja2 and good intentions.
    'flask',

    # Bottle is a fast and simple micro-framework for small web applications.
    'bottle',

    # SQLAlchemy is the Python SQL toolkit and Object Relational Mapper
    # that gives application developers the full power and flexibility of SQL.
    'sqlalchemy',
)


if sys.version_info[0] == 2:
    if platform.python_implementation() != 'PyPy':
        # Test dependencies that don't apply for python 2.x.
        test_dependencies += (
            # gevent is a coroutine-based Python networking library that uses
            # greenlet to provide a high-level synchronous API on top of the
            # libevent event loop.
            'gevent',
        )


setup(
    name='armet',
    version='0.3.11',
    description='Clean and modern framework for creating RESTful APIs.',
    author='Concordus Applications',
    author_email='support@concordusapps.com',
    url='http://github.com/armet/python-armet',
    package_dir={'armet': 'src/armet'},
    packages=find_packages('src'),
    dependency_links=(
        'git+git://github.com/concordusapps/wsgi-intercept.git'
            '#egg=wsgi_intercept-0.6.0',
    ),
    install_requires=(
        # Python 2 and 3 normalization layer
        'six',

        # For parsing accept and content-type headers
        'python-mimeparse',
    ),
    extras_require={
        'test': test_dependencies
    }
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class Deserializer(object):

    #! Applicable media types for this deserializer.
    media_types = ()

    def can_deserialize(self, text=None):
        """Tests this deserializer to see if it can deserialize."""
        try:
            # Attempt to deserialize the object.
            self.deserialize(text)

            # The deserialization process is assumed to have succeed.
            return True

        except ValueError:
            # The object was of an unsupported type.
            return False

    def deserialize(self, text=None):
        """Parses the text into a format consumable by python.

        @throws ValueError
            To indicate this deserializer cannot deserialize the
            passed text.
        """
        return text
